# -*- coding: utf-8 -*-
import datetime
import os
import logging
import time
from os.path import isfile
import shutil
import sys
import ftplib
from ftplib import FTP_TLS
from threading import Lock
import socket
import io
import socks  # socksipy (https://github.com/mikedougherty/SocksiPy)
import typing
import wrapt
from tempfile import TemporaryFile

from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer
from pyftpdlib.authorizers import DummyAuthorizer

from d22d.model import midhardware
from d22d.utils import log_info
from d22d.utils.utils import path_leaf, time_stamp
from d22d.utils.ziputils import makedirs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class FTPSocks(ftplib.FTP):
    def __init__(self, host='', user='', passwd='', acct='',
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT, source_address=None,
                 proxyconfig=None):
        """Like ftplib.FTP constructor, but with an added `proxyconfig` kwarg

        `proxyconfig` should be a dictionary that may contain the following
        keys:

        proxytype - The type of the proxy to be used. Three types
                are supported: PROXY_TYPE_SOCKS4 (including socks4a),
                PROXY_TYPE_SOCKS5 and PROXY_TYPE_HTTP
        addr -      The address of the server (IP or DNS).
        port -      The port of the server. Defaults to 1080 for SOCKS
                servers and 8080 for HTTP proxy servers.
        rdns -      Should DNS queries be preformed on the remote side
                (rather than the local side). The default is True.
                Note: This has no effect with SOCKS4 servers.
        username -  Username to authenticate with to the server.
                The default is no authentication.
        password -  Password to authenticate with to the server.
                Only relevant when username is also provided.
        """
        self.proxyconfig = proxyconfig or {}
        ftplib.FTP.__init__(self, host, user, passwd, acct, timeout, source_address)

    def __str__(self):
        return f"ftp://{self.host}:{self.port} with proxy={self.proxyconfig}"

    def connect(self, host='', port=0, timeout=-999, source_address=None):
        """Connect to host.  Arguments are:
         - host: hostname to connect to (string, default previous host)
         - port: port to connect to (integer, default previous port)
         - timeout: the timeout to set against the ftp socket(s)
         - source_address: a 2-tuple (host, port) for the socket to bind
           to as its source address before connecting.
        """
        if host != '':
            self.host = host
        if port > 0:
            self.port = port
        if timeout != -999:
            self.timeout = timeout
        if source_address is not None:
            self.source_address = source_address
        logger.info(f'正在连接FTP{self}')
        sys.audit("ftplib.connect", self, self.host, self.port)
        self.sock = self.create_connection(self.host, self.port)
        self.af = self.sock.family
        self.file = self.sock.makefile('r', encoding=self.encoding)
        self.welcome = self.getresp()
        return self.welcome

    def create_connection(self, host=None, port=None):
        host, port = host or self.host, port or self.port
        if self.proxyconfig:
            phost, pport = self.proxyconfig['addr'], self.proxyconfig['port']
            err = None
            for res in socket.getaddrinfo(phost, pport, 0, socket.SOCK_STREAM):
                af, socktype, proto, canonname, sa = res
                sock = None
                try:
                    sock = socks.socksocket(af, socktype, proto)
                    sock.setproxy(**self.proxyconfig)

                    if self.timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
                        sock.settimeout(self.timeout)
                    sock.connect((host, port))
                    return sock
                except socket.error as _:
                    err = _
                    if sock is not None:
                        sock.close()
            if err is not None:
                raise err
            else:
                raise socket.error("getaddrinfo returns an empty list")
        else:
            sock = socket.create_connection((host, port), self.timeout)
        return sock

    def ntransfercmd(self, cmd, rest=None):
        """Initiate a transfer over the data connection.

        If the transfer is active, send a port command and the
        transfer command, and accept the connection.  If the server is
        passive, send a pasv command, connect to it, and start the
        transfer command.  Either way, return the socket for the
        connection and the expected size of the transfer.  The
        expected size may be None if it could not be determined.

        Optional `rest' argument can be a string that is sent as the
        argument to a REST command.  This is essentially a server
        marker used to tell the server to skip over any data up to the
        given marker.
        """
        size = None
        if self.passiveserver:
            host, port = self.makepasv()
            conn = self.create_connection(host, port)
            try:
                if rest is not None:
                    self.sendcmd("REST %s" % rest)
                resp = self.sendcmd(cmd)
                # Some servers apparently send a 200 reply to
                # a LIST or STOR command, before the 150 reply
                # (and way before the 226 reply). This seems to
                # be in violation of the protocol (which only allows
                # 1xx or error messages for LIST), so we just discard
                # this response.
                if resp[0] == '2':
                    resp = self.getresp()
                if resp[0] != '1':
                    raise ftplib.error_reply(resp)
            except:
                conn.close()
                raise
        else:
            with self.makeport() as sock:
                if rest is not None:
                    self.sendcmd("REST %s" % rest)
                resp = self.sendcmd(cmd)
                # See above.
                if resp[0] == '2':
                    resp = self.getresp()
                if resp[0] != '1':
                    raise ftplib.error_reply(resp)
                conn, sockaddr = sock.accept()
                if self.timeout is not ftplib._GLOBAL_DEFAULT_TIMEOUT:
                    conn.settimeout(self.timeout)
        if resp[:3] == '150':
            # this is conditional in case we received a 125
            size = ftplib.parse150(resp)
        return conn, size

    # def ntransfercmd(self, cmd, rest=None):
    #     size = None
    #     if self.passiveserver:
    #         host, port = self.makepasv()
    #         conn = self.create_connection(host, port)
    #         try:
    #             if rest is not None:
    #                 self.sendcmd("REST %s" % rest)
    #             resp = self.sendcmd(cmd)
    #             # Some servers apparently send a 200 reply to
    #             # a LIST or STOR command, before the 150 reply
    #             # (and way before the 226 reply). This seems to
    #             # be in violation of the protocol (which only allows
    #             # 1xx or error messages for LIST), so we just discard
    #             # this response.
    #             if resp[0] == '2':
    #                 resp = self.getresp()
    #             if resp[0] != '1':
    #                 raise ftplib.error_reply(resp)
    #         except:
    #             conn.close()
    #             raise
    #     else:
    #         raise Exception("Active transfers not supported")
    #     if resp[:3] == '150':
    #         # this is conditional in case we received a 125
    #         size = ftplib.parse150(resp)
    #     return conn, size


def set_file_log(log):
    # create a file handler
    handler = logging.FileHandler('ftp_log.log')
    handler.setLevel(logging.DEBUG)

    # create a logging format
    formatter = logging.Formatter('[%(asctime)s][%(name)s][%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    # add the handlers to the logger
    log.addHandler(handler)


def to_simple_chinese(string):
    # otherLanguage 变成中文
    item = string.encode('iso-8859-1').decode('gbk')
    return item


def to_latin_language(string):
    # 简体中文变成 latin
    item = string.encode('gbk').decode('iso-8859-1')
    return item


ERR_STR_SETTING = {
    "主动模式只适用于客户端有公网IP或者服务器与客户端在同一个网段下": [
        # "ftplib.error_perm: 500 I won't open a connection to 192.168.18.111 (only to 222.111.333.111)"
        " I won't open a connection to ",
        "(only to",
    ],
    "File exists": [
        "File exists"
    ],
    "Not a directory": [
        "Not a directory"
    ],
    "No such file or directory": [
        "No such file or directory"
    ],
    "Can't check for file existence": [
        # ftplib.error_perm: 550 Can't check for file existence
        "Can't check for file existence"
    ],
    "Can't change directory to": [
        # Can't change directory to 123f: No such file or directory
        "Can't change directory to"
    ],
    "(measured here),": [
        # 出现于被动模式断点续传从ftp下载文件的时候，
        # 第一次以为是客户端retrbinary函数本来是执行二进制命令，带有文件大小参数以后，无法执行带有中文名的文件路径编码导致
        # 第二次以为是客户端发送命令之后，没返回要的结果，返回了上次还未执行完毕的命令返回的结果冲突导致（下载文件还没完成返回了）
        # ftplib.error_reply: 150 0.926 seconds (measured here), 0.54 Mbytes per second
        "(measured here),"
    ],
    "Unknown command": [
        # 出现于被动模式断点续传从ftp下载文件的时候， 服务端不支持断点续传
        "Unknown command"
    ],
}


def get_err_str_setting(err_key, ex):
    return any(err_str in str(ex) for err_str in ERR_STR_SETTING[err_key])


def with_ftp_lock():
    """
    请求self.ftp,自带线程锁
    """

    @wrapt.decorator
    def wrapper(func, instance, args, kwargs):
        res = None
        while True:
            lock = instance.ftp_lock
            if lock.locked():
                raise SystemError(f'FTP客户端函数[{func.__name__}]执行错误 ftp locked')
            lock.acquire()
            try:
                res = func(*args, **kwargs)
                break
            except ftplib.error_perm as inst:
                if get_err_str_setting('主动模式只适用于客户端有公网IP或者服务器与客户端在同一个网段下', inst):
                    logger.error("主动模式只适用于客户端有公网IP或者服务器与客户端在同一个网段下")
                    instance.try_all_connect_to(False)
                logger.error(f"FTP客户端函数[{func.__name__}]执行错误：[{inst}] {args} {kwargs}")
                raise inst
            except NETWORK_ERR as inst:
                logger.error(f"FTP客户端函数[{func.__name__}]执行错误：网络中断导致{instance}服务端中断,错误：[{inst}]，5秒后重试...")
                time.sleep(5)
                instance.connect_until_success()
            except Exception as inst:
                raise inst
            finally:
                lock.release()
        return res

    return wrapper


NETWORK_ERR = (
    socket.timeout,  # 直接断了
    EOFError,  # 用代理断了
    socks.ProxyConnectionError,  # 用代理的时候被拔网线
    OSError  # 被拔网线
)


class FtpController:
    # /!\ Although the comments and variable names say 'file_name'/'file_anything' it inculdes folders also
    # Some functions in this class has no exception handling, it has to be done outside

    def __init__(
            self, host, port=21, username=' ', password=' ', use_tls=False, pasv=True, encoding='utf-8',
            proxy_config=None
    ):
        # List to store file search and search keywords
        self.search_file_list = []
        self.detailed_search_file_list = []
        self.keyword_list = []

        # Variable to hold the max no character name in file list (used for padding in GUIs)
        self.max_len = 0

        self.max_len_name = ''

        # Variable to tell weather hidden files are enabled
        self.hidden_files = False

        # Variable to store the platform the server is running on
        self.server_platform = 'Linux' if sys.platform != 'win32' else sys.platform
        self.lock = Lock()
        self.ftp_lock = Lock()
        self.ftp = None
        self.remote_directory_path = '/'
        self.work_dir_now = '/'
        self.filesMoved = 0
        self.bytes_downloaded = 0
        self.bytes_uploaded = 0
        self.up_file_size_start = 0
        self.last_log = 0

        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.use_tls = use_tls
        self.pasv = pasv
        self.encoding = encoding
        self.encoding_test = None
        self.proxyconfig = proxy_config

    def __str__(self):
        return f"ftp://{self.host}:{self.port}{self.work_dir_now}"

    def connect_until_success(self, retry=0):
        cnt_retry = 0
        while True:
            cnt_retry += 1
            try:
                log_info(f"正在尝试第[{cnt_retry}/{retry}]次登录[{self}]。。。")
                # 先无tls，在试试有tls
                self.ftp = None
                if not self.use_tls:
                    self.ftp = FTPSocks(
                        timeout=25, proxyconfig=self.proxyconfig)
                    self.ftp.connect(self.host, self.port)
                else:
                    self.ftp = FTP_TLS(timeout=25)
                    self.ftp.connect(self.host, self.port)
                self.ftp.login(self.username, self.password)

                if self.use_tls:
                    self.ftp.prot_p()
                self.ftp.cwd(self.work_dir_now)
                if self.pasv is not None:
                    self.ftp.set_pasv(self.pasv)
                self.ftp.encoding = self.encoding
                logger.info(f"Connected. now pwd: {self.work_dir_now}")
                return
            except Exception as inst:
                if retry and cnt_retry > retry:
                    raise inst
                print_func = logger.error
                if not isinstance(inst, NETWORK_ERR):
                    print_func = logger.exception
                print_func(f"第[{cnt_retry}/{retry}]次登录[{self}]失败 [{type(inst)}] {inst}, 5秒后重试。。。")
                time.sleep(5)

    def try_all_connect_to(self, no_pasv=True):
        pasv_d = {
            True: "被动模式",
            False: "主动模式",
            None: "默认模式"  # 如果 val 为 true，则打开“被动”模式，否则禁用被动模式。默认下被动模式是打开的。主动模式只适用于客户端有公网IP或者服务器在同一个网段下

        }
        for tls, pasv in [
                             (True, None),
                             (False, None),
                             (True, True),
                             (False, True),
                         ] + ([
            (True, False),
            (False, False),
        ] if no_pasv else []):
            if self.try_connect_to(tls, pasv):
                logger.info(f'成功! 登录尝试FTP：{"使用" if tls else "无"}tls {pasv_d[pasv]}')
                return tls, pasv
            else:
                logger.info(f'尝试登录FTP失败：{"使用" if tls else "无"}tls {pasv_d[pasv]}')
        raise SystemError(f'尝试登录FTP全部失败{self}')

    def try_connect_to(self, use_tls, pasv, encoding="utf-8"):
        try:
            # 先无tls，在试试有tls
            self.ftp = None
            if not use_tls:
                self.ftp = FTPSocks(
                    timeout=25, proxyconfig=self.proxyconfig)
                self.ftp.connect(self.host, self.port)
            else:
                self.use_tls = True
                self.ftp = FTP_TLS(timeout=3)
                self.ftp.connect(self.host, self.port)
            self.ftp.login(self.username, self.password)

            if use_tls:
                self.ftp.prot_p()
            self.ftp.cwd(self.remote_directory_path)
            if pasv is not None:
                self.ftp.set_pasv(pasv)
            self.ftp.encoding = encoding
            logger.info(f"Connected. now pwd: {self.remote_directory_path}")
            self.use_tls = use_tls
            self.pasv = pasv
            self.encoding = encoding
            return True
        except NETWORK_ERR as inst:
            logger.error(f"重试连接{self}服务端中断,错误：[{inst}]")
        except Exception as inst:
            logger.exception(f"Couldn't connect server: {type(inst)}")
            return False

    def check_encoding(self, encoding="utf-8"):
        logger.info(f"正在测试到服务端encoding值")
        old_encoding = encoding
        self.ftp.encoding = encoding
        self.encoding_test = None
        test_string = '测试用Test，空文件夹，看到请顺手删除，请勿往里面存数据，Please_delete_handy!'
        # TODO 新建中文名来测试文件夹名是否能取到
        self.make_dir_optimistic(test_string)
        for file in self.get_detailed_file_list():
            if test_string in file:
                self.encoding = encoding
                self.encoding_test = encoding
                logger.info(f"正确测试到服务端encoding值为：{encoding}")
                break
        if not self.encoding_test:
            logger.info(f"测试到服务端encoding值不为：{encoding}，请更换")
            self.ftp.encoding = old_encoding
            self.encoding_test = None
        self.delete_dir_fast(test_string)

    def disconnect(self):
        self.ftp.quit()  # Close FTP connection
        self.ftp = None

    def toggle_hidden_files(self):
        self.hidden_files = not self.hidden_files

    @with_ftp_lock()
    def get_detailed_file_list(self, ignore_hidden_files_flag=False):
        files = []

        def dir_callback(line):
            if self.server_platform != 'Linux':
                files.append(line)
                return
            if (self.hidden_files is True or line.split()[8][0] != '.') or ignore_hidden_files_flag is True:
                files.append(line)

        self.ftp.dir(dir_callback)
        return files

    def get_file_list(self, detailed_file_list):
        self.max_len = 0
        self.max_len_name = ''
        file_list = []
        for x in detailed_file_list:
            # Remove details and append only the file name
            if self.server_platform == 'Linux':
                name = self.get_properties(x)[0]
            else:
                name = x
            file_list.append(name)
            if len(name) > self.max_len:
                self.max_len = len(name)
                self.max_len_name = name
        return file_list

    def get_detailed_search_file_list(self):
        return self.detailed_search_file_list

    def get_search_file_list(self):
        self.max_len = 0
        self.max_len_name = ''
        for name in self.search_file_list:
            if len(name) > self.max_len:
                self.max_len = len(name)
                self.max_len_name = name
        return self.search_file_list

    def chmod(self, filename, permissions):
        self.ftp.sendcmd('SITE CHMOD ' + str(permissions) + ' ' + filename)

    def is_there(self, path):
        try:
            self.ftp.sendcmd('MLST ' + path)
            return True
        except:
            return False

    def rename_dir(self, rename_from, rename_to):
        self.ftp.sendcmd('RNFR ' + rename_from)
        self.ftp.sendcmd('RNTO ' + rename_to)

    def move_dir(self, rename_from, rename_to, status_command=log_info, replace_command=log_info):
        if self.is_there(rename_to) is True:
            if replace_command(rename_from, 'File/Folder exists in destination folder') is True:
                self.delete_dir(rename_to, status_command)
            else:
                return
        try:
            self.ftp.sendcmd('RNFR ' + rename_from)
            self.ftp.sendcmd('RNTO ' + rename_to)
            status_command(rename_from, 'Moved')
        except:
            status_command(rename_from, 'Failed to move')

    def copy_file(self, file_dir, copy_from, file_size, status_command=log_info, replace_command=log_info):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.pwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_file(copy_from, file_size, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_file(copy_from, file_size, status_command, replace_command)
        # Delete the downloaded file
        os.remove(copy_from)
        status_command(copy_from, 'Deleted local file')

    def copy_dir(self, file_dir, copy_from, status_command=log_info, replace_command=log_info):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.pwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_dir(copy_from, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_dir(copy_from, status_command, replace_command)
        # Delete the downloaded folder
        shutil.rmtree(copy_from)
        status_command(copy_from, 'Deleting local directory')

    def delete_file(self, file_name, status_command=log_info):
        try:
            self.ftp.sendcmd('DELE ' + file_name)
            status_command(file_name, 'Deleted')
        except:
            status_command(file_name, 'Failed to delete')

    def delete_dir(self, dir_name, status_command=log_info):
        # Go into the directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if self.is_dir(file_details):
                self.delete_dir(file_name, status_command)
            # If file
            else:
                self.delete_file(file_name, status_command)
        # Go back to parent directory and delete it
        try:
            self.ftp.cwd('..')
            status_command(dir_name, 'Deleting directory')
            self.ftp.sendcmd('RMD ' + dir_name)
        except:
            status_command(dir_name, 'Failed to delete directory')
            return

    @with_ftp_lock()
    def delete_dir_fast(self, dir_name, status_command=log_info):
        status_command(dir_name, '正在删除文件夹，Deleting directory')
        self.ftp.sendcmd('RMD ' + dir_name)

    def upload_file(self, file_name, file_size, status_command=log_info, replace_command=log_info):
        def update_progress(data):
            self.bytes_uploaded += int(sys.getsizeof(data))
            status_command(file_name, str(min(round((self.bytes_uploaded / file_size) * 100, 8), 100)) + '%')

        # Variable to keep trak of number of bytes uploaded
        self.bytes_uploaded = 0
        # Check if the file is already present in ftp server
        if self.is_there(file_name):
            if replace_command(file_name, 'File exists in destination folder') is False:
                return
        # Try to open file, if fails return
        try:
            file_to_up = open(file_name, 'rb')
        except:
            status_command(file_name, 'Failed to open file')
            return
        # Try to upload file
        try:
            status_command(file_name, 'Uploading')
            self.ftp.storbinary('STOR ' + file_name, file_to_up, 8192, update_progress)
            status_command(None, 'newline')
        except:
            status_command(file_name, 'Upload failed')
            return
        # Close file
        file_to_up.close()

    @staticmethod
    def format_realpath(path):
        if sys.platform == 'win32':
            res = str(os.path.realpath(path)).split(':', 1)[-1].replace('\\', '/')
        else:
            res = os.path.realpath(path)
        return res

    @staticmethod
    def format_path(path):
        if sys.platform == 'win32':
            res = str(path).replace('\\', '/')
        else:
            res = path
        return res

    def upload_file_to_some_where(
            self, local_path, remote_folder, remote_filename='', status_command=log_info, check_ftp_file_same=True):
        # TODO BytesIO类型local_path
        if not os.path.exists(local_path):
            raise SystemError(f'本地路径不存在：{local_path.__repr__()}')
        if not remote_folder:
            raise SystemError(f'远程路径错误：{remote_folder.__repr__()} {remote_filename}')
        if not remote_filename:
            remote_folder, remote_filename = os.path.split(remote_folder)

        if not remote_folder.startswith('/'):
            remote_folder = self.format_realpath(os.path.join(self.work_dir_now, remote_folder))
        remote_path = os.path.join(remote_folder, remote_filename)
        old_path = self.work_dir_now
        if remote_folder:
            try:
                self.cwd_recode_path(remote_folder)
            except ftplib.error_perm as ex:
                if get_err_str_setting("Can't change directory to", ex):
                    self.make_dir_optimistic(remote_folder)
        self._upload_file_to_some_where(
            local_path, remote_path, remote_filename, status_command, check_ftp_file_same)
        self.work_dir_now = old_path
        self.cwd_recode_path(old_path)

    @with_ftp_lock()
    def _upload_file_to_some_where(
            self, local_path, remote_path, remote_filename, status_command=log_info, check_ftp_file_same=True):
        file_size = os.stat(local_path).st_size
        try:
            if check_ftp_file_same:
                status_command('正在检查远程已经存在的文件路径上传的文件大小')
                self.up_file_size_start = self.get_size(remote_path) or 0

                _all_size = f'{file_size / 1024 / 1024:.3f}MB'
                _up_size = f'{self.up_file_size_start / 1024 / 1024:.3f}MB'
                status_command(f"服务端已存在同名文件的文件大小:{_all_size},本地{_up_size}")

                def check_file(data):
                    f_data = file_need_up.read(len(data))
                    status_command(
                        f"正在检查远程ftp服务器已经存在的文件路径上传的文件和本地文件一致性：\n",
                        f"本地：{f_data[:70]}\n远程：{data[:70]}\n本地{len(f_data)}=?远程{len(data)}: {f_data == data}")
                    if f_data == data:
                        self.is_same_file = True
                        status_command(f"文件开头800KB一致，准备开始断点续传，已经上传的文件大小:{self.up_file_size_start / 1024 / 1024:.3f}MB", )
                    else:
                        self.is_same_file = False
                        raise SystemError(f'远程文件路径["{self}{remote_path}:1"]已存在，而且本地文件开头800KB与服务器文件不一致，请检查')
                    raise StopIteration('只是检查文件开头一致性，不需要全部下载')

                with open(local_path, 'rb') as file_need_up:
                    status_command(f'正在下载远程服务器文件[{file_size / 1024 / 1024:.3f}MB]的开头800KB进行文件比对，路径：{remote_path}')
                    try:
                        self.ftp.retrbinary('RETR ' + remote_path, check_file, blocksize=1024 * 8 * 128)  # 检查800KB
                    except StopIteration:
                        try:
                            status_command(self.ftp.getmultiline())
                            self.ftp.abort()
                        except NETWORK_ERR:
                            self.ftp.close()
                            self.connect_until_success()

            else:
                self.up_file_size_start = 0
        except ftplib.error_perm as ex:
            if get_err_str_setting("Can't check for file existence", ex):
                self.up_file_size_start = 0
            else:
                raise
        start_time = time.time()
        self.bytes_uploaded = 0
        self.last_log = time.time()

        def upload_file(data):
            self.bytes_uploaded += int(sys.getsizeof(data))
            all_size = f'{file_size / 1024 / 1024:.3f}MB'
            up_size = f'{(self.up_file_size_start + self.bytes_uploaded) / 1024 / 1024:.3f}MB'
            per_second_size = f'{self.bytes_uploaded / 1024 / 1024 / (time.time() - start_time):.3f}MB/s'
            if self.last_log + 2 < (time_now := time.time()):
                self.last_log = time_now
                status_command(remote_path, str(min(
                    round(((self.up_file_size_start + self.bytes_uploaded) / file_size) * 100, 3),
                    100)) + f'% [{up_size}/ {all_size}] {per_second_size}')

        if self.up_file_size_start == file_size:
            return
        elif self.up_file_size_start < file_size:
            with open(local_path, 'rb') as local_file:
                local_file.seek(self.up_file_size_start)
                self.ftp.storbinary(
                    'STOR ' + remote_filename, local_file,
                    blocksize=8192, callback=upload_file, rest=self.up_file_size_start)
        else:
            raise SystemError(f'本地文件大小 小于 远程文件大小：{local_path} --> {self}:{remote_path}')

    @staticmethod
    def get_split_dir_path_list(path):
        res = [(path, '')]
        while True:
            root_path, final_dir = os.path.split(path)
            res.append((root_path, final_dir))
            path = root_path
            if root_path == "" or root_path == '/' or root_path.endswith(':\\'):
                break
        return res

    @with_ftp_lock()
    def make_dir_optimistic(self, path_dir):
        """
        乐观创建文件夹，假定所有父文件都已经被创建，此时效率高，能直接越级创建深层级文件夹。
        :return:
        """

        path_dir_l = self.get_split_dir_path_list(path_dir)
        need_make = []
        logger.info(f'乐观创建文件夹: "{os.path.join(self.work_dir_now, path_dir)}"')
        for root, folder in path_dir_l:
            try:
                if root:
                    self.ftp.mkd(root)
            except (ftplib.error_perm,) as inst:
                # print(root, str(ex))
                if get_err_str_setting('No such file or directory', inst):
                    need_make.insert(0, [root, folder])
                elif get_err_str_setting("File exists", inst):
                    break
                elif get_err_str_setting("Not a directory", inst):
                    raise SystemError(f'创建{self}文件夹时出现同名文件：{root}') from None
                else:
                    raise SystemError(f'创建{self}文件夹时出现问题：{root} {inst}')

        for root, folder in need_make:
            try:
                self.ftp.mkd(root)
            except (ftplib.error_perm,) as ex:
                # print(root, str(ex))
                raise SystemError(f'创建{self}文件夹时出现问题：{root} {ex}')

    def make_dir_pessimistic(self, path_dir):
        """
        悲观创建文件夹，假定文件都没被创建，此时效率高，能直接越级创建深层级文件夹。
        :return:
        """
        path_dir_l = self.get_split_dir_path_list(path_dir)
        path_dir_l.reverse()
        need_make = []
        for root, folder in path_dir_l:
            try:
                if root:
                    self.ftp.mkd(root)
            except (ftplib.error_perm,) as inst:
                # print(root, str(ex))
                if get_err_str_setting('File exists', inst):
                    continue
                elif get_err_str_setting("Not a directory", inst):
                    raise SystemError(f'创建{self}文件夹时出现同名文件：{root}') from None
                else:
                    raise SystemError(f'创建{self}文件夹时出现问题：{root} {inst}')

        for root, folder in need_make:
            self.ftp.mkd(root)

    def upload_dir(self, dir_name, status_command=log_info, replace_command=log_info):
        # Change to directory
        os.chdir(dir_name)
        # Create directory in server and go inside
        try:
            if not self.is_there(dir_name):
                self.ftp.mkd(dir_name)
                status_command(dir_name, 'Creating directory')
            else:
                status_command(dir_name, 'Directory exists')
            self.ftp.cwd(dir_name)
        except:
            status_command(dir_name, 'Failed to create directory')
            return
        # Cycle through items
        for filename in os.listdir():
            # If file upload
            if isfile(filename):
                self.upload_file(filename, os.path.getsize(filename), status_command, replace_command)
            # If directory, recursive upload it
            else:
                self.upload_dir(filename, status_command, replace_command)

        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def get_size_etc(self, ftp_file_name):
        old_path = self.work_dir_now
        path, _file_name = os.path.split(ftp_file_name)
        if path:
            self.cwd_recode_path(path)
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        res = 1
        for file_name, file_details in zip(file_list, detailed_file_list):
            if file_name == _file_name and not self.is_dir(file_details):
                res = int(self.get_properties(file_details)[3])
        if path:
            self.cwd_recode_path(old_path)
        return res

    def walk(self, ftp_file_path):
        old_path = self.work_dir_now
        if ftp_file_path:
            if ftp_file_path.startswith('/'):
                ftp_file_path = self.format_realpath(ftp_file_path)
            else:
                ftp_file_path = self.format_realpath(os.path.join(self.work_dir_now, ftp_file_path))
            self.cwd_recode_path(ftp_file_path)
        else:
            raise FileNotFoundError(ftp_file_path)
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            r_fs = []
            r_fns = []
            pr = self.get_properties(file_details)
            file_name = pr[0]
            if file_name and file_name in ['.', '..']:
                continue
            if self.is_dir(file_details):
                r_fs.append(file_name)
                for root, fs, fns in self.walk(f"{ftp_file_path}/{file_name}"):
                    yield root, fs, fns
            else:
                file_attribs, date_modified = pr[1], pr[2]
                r_fns.append((file_name, file_attribs, date_modified, int(pr[-1] if pr[-1] else 0)))
            yield ftp_file_path, r_fs, r_fns

        if ftp_file_path:
            self.cwd_recode_path(old_path)

    def get_size(self, ftp_file_name):
        res = int(self.ftp.size(ftp_file_name)) or 0
        return res

    @with_ftp_lock()
    def get_size_until_success(self, ftp_file_name):
        res = int(self.ftp.size(ftp_file_name)) or 0
        return res

    def download_file_to_some_where(
            self, ftp_file_name, local_path, local_file_name='',
            file_size=0, status_command=log_info, replace_command=log_info, check_ftp_file_same=False
    ):
        if not local_path:
            raise SystemError(f'路径错误：{local_path.__repr__()}')
        # Variable to keep track of total bytes downloaded
        # Check if the file is already present in local directory
        if isfile(ftp_file_name):
            if replace_command(ftp_file_name, 'File exists in destination folder') is False:
                return
        # Try to open file, if fails return
        if not local_file_name:
            local_file_name = os.path.basename(ftp_file_name)

        makedirs(local_path, check_dot=False)
        local_path = os.path.join(local_path, local_file_name)

        # if local_path.endswith('/'):
        #     makedirs(local_path)
        #     local_path = os.path.join(local_path, local_file_name)
        # else:
        #     makedirs(os.path.dirname(local_path))

        if not file_size:
            file_size = self.get_size_until_success(ftp_file_name)

        self._download_file_to_some_where(
            ftp_file_name, local_path, file_size, status_command, check_ftp_file_same=check_ftp_file_same)
        return local_path

    @with_ftp_lock()
    def _download_file_to_some_where(
            self, ftp_file_name, local_path,
            file_size, status_command=log_info, check_ftp_file_same=False
    ):
        self.bytes_downloaded = 0
        self.is_same_file = False
        self.down_file_size_start = 0

        self.last_log = 0

        # Function for updating status and writing to file

        def write_file(data):
            self.bytes_downloaded += int(sys.getsizeof(data))
            all_size = f'{file_size / 1024 / 1024:.3f}MB'
            down_size = f'{(self.down_file_size_start + self.bytes_downloaded) / 1024 / 1024:.3f}MB'
            per_second_size = f'{self.bytes_downloaded / 1024 / 1024 / (time.time() - start_time):.3f}MB/s'
            if self.last_log + 2 < (time_now := time.time()):
                self.last_log = time_now
                status_command(ftp_file_name,
                               str(min(
                                   round(((self.down_file_size_start + self.bytes_downloaded) / file_size) * 100, 3),
                                   100)) + f'% [{down_size}/ {all_size}] {per_second_size}')
            file_to_down.write(data)

        def check_file(data):
            bytes_downloaded = int(sys.getsizeof(data))
            all_size = f'{file_size / 1024 / 1024:.3f}MB'
            down_size = f'{bytes_downloaded / 1024 / 1024:.3f}MB'
            per_second_size = f'{bytes_downloaded / 1024 / 1024 / (time.time() - start_time):.3f}MB/s'
            status_command(
                '正在检查本地已经存在的文件路径下载的文件大小', ftp_file_name, str(min(round(
                    (bytes_downloaded / file_size) * 100, 3), 100)) + f'% [{down_size}/ {all_size}] {per_second_size}')

            f_data = file_need_down.read(len(data))
            status_command(f"比较服务器和本地文件一致性：\n{f_data[:200]}\n{data[:200]}\n{len(f_data)}=?{len(data)}:{f_data == data}")
            if f_data == data:
                self.is_same_file = True
                self.down_file_size_start = os.path.getsize(local_path)
                status_command(f"已经下载的文件大小:{self.down_file_size_start / 1024 / 1024:.3f}MB", )
            else:
                self.is_same_file = False
                raise SystemError(f'本地文件路径["{local_path}:1"]已存在，而且文件开头与服务器不一致，请检查')
            raise StopIteration('只是检查文件开头一致性，不需要全部下载')

        start_time = time.time()
        if os.path.exists(local_path) and os.path.getsize(local_path):
            if file_size and file_size == os.path.getsize(local_path):
                status_command(f"已经下载的文件大小 与 传入的文件大小参数 一致，跳过下载:{file_size / 1024 / 1024:.3f}MB", )
                return
            if check_ftp_file_same:
                with open(local_path, 'rb') as file_need_down:
                    status_command(
                        f'正在下载文件 [{file_size / 1024 / 1024:.3f}MB] {ftp_file_name} --> {os.path.realpath(local_path)}')
                    try:
                        self.ftp.retrbinary('RETR ' + ftp_file_name, check_file, blocksize=1024 * 8 * 128)  # 检查800KB
                    except StopIteration:
                        try:
                            status_command(self.ftp.getmultiline())
                            self.ftp.abort()
                        except NETWORK_ERR:
                            self.ftp.close()
                            self.connect_until_success()
            else:
                self.is_same_file = True
                self.down_file_size_start = os.path.getsize(local_path)

        start_time = time.time()
        with open(local_path, 'ab') as file_to_down:
            status_command(ftp_file_name,
                           f'Downloading file[{self.down_file_size_start / 1024 / 1024}MB/{file_size / 1024 / 1024:.3f}MB] to {os.path.realpath(local_path)}')
            self.ftp.retrbinary('RETR ' + ftp_file_name, write_file, rest=self.down_file_size_start or None)
            status_command(ftp_file_name,
                           f'Download success [{time.time() - start_time}s {file_size / 1024 / 1024:.3f}MB] to {os.path.realpath(local_path)}')

    def download_file(self, ftp_file_name, file_size=0, status_command=log_info, replace_command=log_info):
        # Function for updating status and writing to file
        def write_file(data):
            self.bytes_downloaded += int(sys.getsizeof(data))
            status_command(ftp_file_name, str(min(round((self.bytes_downloaded / file_size) * 100, 3), 100)) + '%')
            file_to_down.write(data)

        # Variable to keep track of total bytes downloaded
        self.bytes_downloaded = 0
        # Check if the file is already present in local directory
        if isfile(ftp_file_name):
            if replace_command(ftp_file_name, 'File exists in destination folder') is False:
                return
        # Try to open file, if fails return
        try:
            file_to_down = open(ftp_file_name, 'wb')
        except:
            status_command(ftp_file_name, 'Failed to create file')
            return
        # Try to upload file
        try:
            status_command(ftp_file_name, 'Downloading')
            self.ftp.retrbinary('RETR ' + ftp_file_name, write_file)
            status_command(None, 'newline')
        except:
            status_command(ftp_file_name, 'Download failed')
        # Close file
        file_to_down.close()

    def download_dir(self, ftp_dir_name, status_command=log_info, replace_command=log_info):
        # Create local directory
        try:
            if not os.path.isdir(ftp_dir_name):
                os.makedirs(ftp_dir_name)
                status_command(ftp_dir_name, 'Created local directory')
            else:
                status_command(ftp_dir_name, 'Local directory exists')
            os.chdir(ftp_dir_name)
        except:
            status_command(ftp_dir_name, 'Failed to create local directory')
            return
        # Go into the ftp directory
        self.ftp.cwd(ftp_dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if self.is_dir(file_details):
                self.download_dir(file_name, status_command, replace_command)
            # If file
            else:
                self.download_file(file_name, int(self.get_properties(file_details)[3]), status_command,
                                   replace_command)
        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def search(self, dir_name, search_file_name, status_command=log_info):
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If file_name matches the keyword, append it to search list
            if search_file_name.lower() in file_name.lower():
                if self.ftp.pwd() == '/':
                    dir_ = ''
                else:
                    dir_ = self.ftp.pwd()
                self.search_file_list.append(dir_ + '/' + file_name)
                self.detailed_search_file_list.append(file_details)
                status_command(dir_ + '/' + file_name, 'Found')
                # If directory, search it
            if self.is_dir(file_details):
                status_command(file_name, 'Searching directory')
                self.search(file_name, search_file_name, status_command)
        # Goto to parent directory
        self.ftp.cwd('..')

    def clear_search_list(self):
        del self.search_file_list[:]
        del self.detailed_search_file_list[:]

    def get_dir_size(self, dir_name):
        size = 0
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            if self.is_dir(file_details):
                size += self.get_dir_size(file_name)
            else:
                size += int(self.get_properties(file_details)[3])
        # Goto to parent directory
        self.ftp.cwd('..')
        return size

    def cwd_parent(self, name):
        if '/' not in name:
            return name
        parent_name = '/'.join(name.split('/')[:-1])
        if parent_name == '':
            parent_name = '/'
        self.ftp.cwd(parent_name)
        return ''.join(name.split('/')[-1:])

    def mkd(self, name):
        self.ftp.mkd(name)

    def pwd(self):
        return self.ftp.pwd()

    def get_properties(self, file_details):
        if self.server_platform == 'Linux':
            details_list = file_details.split()
            # Get file attributes
            file_attribs = details_list[0]
            # Get date modified
            date_modified = ' '.join(details_list[5:8])
            # Remove the path from the name
            file_name = ' '.join(details_list[8:])
            # Get size if it is not a directory
            if 'd' not in file_details[0]:
                file_size = details_list[4]
                return [file_name, file_attribs, date_modified, file_size]
            else:
                return [file_name, file_attribs, date_modified]

    def is_dir(self, file_details):
        if self.server_platform == 'Linux':
            return 'd' in file_details[0]

    @with_ftp_lock()
    def cwd_recode_path(self, path):
        self.ftp.cwd(path)
        if path.startswith('/'):
            self.work_dir_now = self.format_realpath(path)
        else:
            old_path = self.work_dir_now
            self.work_dir_now = self.format_realpath(os.path.join(old_path, path))

    def moveFTPFiles(self, fileToUpload, remoteDirectoryPath):
        self.lock.acquire(True)
        try:
            localFileName = path_leaf(fileToUpload)
            if localFileName != "":
                # create a copy of the file before sending it
                # tempFile = create_temporary_copy(fileToUpload, workingDir, localFileName)
                # open a the local file
                with open(fileToUpload, 'rb') as rf:
                    wf = TemporaryFile('w+b')
                    while datar := rf.read(8128):
                        wf.write(datar)
                    wf.seek(0)
                    # Download the file a chunk at a time using RETR

                    # p_start = self.pwd()
                    p_start = self.work_dir_now
                    self.cwd(remoteDirectoryPath)
                    self.ftp.storbinary('STOR ' + localFileName, wf)
                    self.cwd(p_start)

                    wf.close()
                logger.info("Uploaded file: " + str(fileToUpload) + " on " + time_stamp())

                # Close the file
                # remove the temp file
                # to remove it I need to have the right permissions
                # os.chmod(tempFile, 777)
                # os.remove(tempFile)
        except Exception as inst:
            logger.exception(f"Connection Error - {inst}" + time_stamp())

        self.lock.release()


class FtpClientFileStore(midhardware.BaseStore):
    # TODO
    pass


class FtpClientStore(midhardware.BaseStore):
    def __init__(self, host, port, username, password, location='/', tmp_path='ftp_data_tmp',
                 use_tls=False, pasv=None, encoding='utf-8', socks_proxy: typing.Union[str, tuple, dict] = None,
                 download_check_ftp_file_same=False, upload_check_ftp_file_same=True):
        if socks_proxy:
            if isinstance(socks_proxy, str):
                p_host, p_port = socks_proxy.split(':')
            elif isinstance(socks_proxy, tuple):
                p_host, p_port = socks_proxy
            else:
                p_host, p_port = socks_proxy.get('p_host'), socks_proxy.get('p_port')
            socks_proxy = dict(proxytype=socks.PROXY_TYPE_SOCKS5, rdns=True, addr=p_host, port=int(p_port))

        self.client = FtpController(host, port, username, password, use_tls, pasv, encoding, socks_proxy)
        self.client.connect_until_success()
        self.location = location
        self.download_check_ftp_file_same = download_check_ftp_file_same
        self.upload_check_ftp_file_same = upload_check_ftp_file_same
        self.tmp_path = tmp_path
        self.client.cwd_recode_path(self.location)

    def count_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def list_data(self, data_type=None, location=None, *args, **kwargs):
        for root, fs, fns in self.client.walk(location or self.location):
            for fn, file_attribs, date_modified, size in fns:
                yield {
                    'root': root,
                    'filename': fn,
                    'attribs': file_attribs,
                    'modified': date_modified,
                    'size': int(size),
                    "realpath": self.client.format_realpath(os.path.join(root, fn))
                }

    def check_data(self, position: typing.Union[str, dict, FtpClientFileStore], data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_data(self, position: typing.Union[str, dict, FtpClientFileStore], data_type=None, *args, **kwargs):
        """

        self.tmp_path 本地文件夹路径
        本地文件名不用也不能传 TODO 可选重命名
        position 远程文件名
        self.location 远程文件夹路径
        """
        file_name = None
        file_size = 0
        if isinstance(position, str):
            file_name = self.client.format_path(os.path.join(self.location, position))
        elif isinstance(position, dict):
            file_name = position['realpath']
            file_size = position['size']
        elif isinstance(position, FtpClientFileStore):
            pass
        if file_name:
            return self.client.download_file_to_some_where(
                file_name,
                self.tmp_path,
                file_size=file_size,
                check_ftp_file_same=self.download_check_ftp_file_same
            )

    def save_data(self, position: str, data: typing.Union[str, bytes, io.BytesIO], data_type=None, *args, **kwargs):
        if isinstance(data, str):
            data = data
        elif isinstance(data, (bytes, io.BytesIO)):
            raise NotImplementedError
        else:
            raise SystemError(f'data type error: {type(data)}')
        return self.client.upload_file_to_some_where(
                data,
                self.location,
                position,
                check_ftp_file_same=self.upload_check_ftp_file_same
            )

    def delete_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_position(self, position, data_type=None, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def check_self(self, *args, **kwargs):
        return NotImplementedError

    def save_self(self, *args, **kwargs):
        return NotImplementedError

    def free_self(self, *args, **kwargs):
        return NotImplementedError


class FtpSeverStore(midhardware.BaseStore):
    def __init__(
            self, host='0.0.0.0', port=21, username='user', password="123qwe", location='/home/user',
            use_tls=False, pasv=True, encoding='utf-8'):
        authorizer = DummyAuthorizer()
        authorizer.add_user(username, password, location, perm=password)
        handler = FTPHandler
        handler.authorizer = authorizer

        server = FTPServer((host, port), handler)
        server.serve_forever()
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.location = location

    def count_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def list_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def save_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def delete_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_position(self, position, data_type=None, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def check_self(self, *args, **kwargs):
        return NotImplementedError

    def save_self(self, *args, **kwargs):
        return NotImplementedError

    def free_self(self, *args, **kwargs):
        return NotImplementedError


if __name__ == '__main__':
    __fs = FtpClientStore('127.0.0.1', 21111, 'python', '123456', '/',
                          use_tls=False, pasv=True, encoding='utf-8', socks_proxy=None)
    __fs.client.try_all_connect_to()
    __fs.client.connect_until_success()
    __fs.client.ftp.cwd("123f")
    __fs.client.make_dir_optimistic('/testss1/test1/tests11/wfwe/rdg24/asdfg')
    __fs.client.check_encoding()

    # for __f in __fs.client.get_detailed_file_list():
    #     print(__f)



#!/usr/bin/env python3
# coding: utf-8


import os
import random
from . import logger, db, diskcacheofsqlite
from .utils import run_task_auto_retry, task_auto_retry, async_task_auto_retry, get_file_md5, get_md5, log_info
from .utils import set_shell_log, active_log
from .db import *
from .decorators import print_hz_async, print_hz
from .logger import LogFormatter
from .fetch_to_requests import gen_from_clipboard, fetch_to_requests
from . import f2r


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]



import platform
import os

if platform.system() == 'Windows':
    import win32api
    import win32con


# Function to get all removable mountpoints attached to the computer
def get_mounts():
    mountpoints = []

    if platform.system() == 'Linux':
        f = open('/proc/mounts')

        dev_types = ['/dev/sda', '/dev/sdc', '/dev/sdb', '/dev/hda', '/dev/hdc', '/dev/hdb', '/dev/nvme']

        for line in f:
            details = line.split()
            if details[0][:-1] in dev_types:
                if details[1] != '/boot/efi':
                    details_decoded_string = bytes(details[1], "utf-8").decode("unicode_escape")
                    mountpoints.append(details_decoded_string)

        f.close()
    elif platform.system() == 'Darwin':
        for mountpoint in os.listdir('/Volumes/'):
            mountpoints.append('/Volumes/' + mountpoint)

    elif platform.system() == 'Windows':
        mountpoints = win32api.GetLogicalDriveStrings()
        mountpoints = mountpoints.split('\000')[:-1]

    return mountpoints



# coding=utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sqlparse
import csv
import re
from sqlparse.sql import Identifier, IdentifierList

from sqlparse.tokens import Keyword, Name

RESULT_OPERATIONS = {'UNION', 'INTERSECT', 'EXCEPT', 'SELECT'}
ON_KEYWORD = 'ON'
PRECEDES_TABLE_NAME = {'FROM', 'JOIN', 'DESC', 'DESCRIBE', 'WITH'}


class BaseExtractor(object):
    def __init__(self, sql_statement):

        # TODO format速度慢一倍
        # self.sql = sqlparse.format(sql_statement, reindent=True, keyword_case='upper')
        self.sql = sql_statement
        self._table_names = set()
        self._alias_names = set()
        self._limit = None

        import time
        time_start = time.time()
        self.parsed = sqlparse.parse(self.stripped())
        # for statement in self.parsed:
        #     self.__extract_from_token(statement)
        #     self._limit = self._extract_limit_from_query(statement)
        self._table_names = self._table_names - self._alias_names
        self.sql_type = None
        self.table_name = None
        self.keys = []
        self.values = []

    def get_v_main(self, keys=None):
        import copy
        # self.get_v(self.parsed[0])
        self.get_v_v2(self.parsed[0])
        data = []
        ks = keys or self.keys
        len_k = len(ks)
        # len_v = len(self.values)
        # print(len_k, len_v, ks)
        # print(self.values)
        d_tmp = dict()
        for idx, v in enumerate(self.values):
            if v['type'] == 'Integer':
                res = int(v['value'])
            # if v['type'] == 'Single':
            #     res = v['value']
            elif v['type'] == 'Keyword' and v['value'] == 'NULL':
                res = None
            else:
                val = v['value']
                if val.startswith("'") or val.startswith('"'):
                    val = val[1:]
                if val.endswith("'") or val.endswith('"'):
                    val = val[:-1]
                res = val

            d_tmp[ks[idx % len_k]] = res
            if d_tmp and idx and not (idx + 1) % len_k:
                data.append(copy.deepcopy(d_tmp))
                d_tmp = dict()
        # print(data)
        return data

    @property
    def tables(self):
        return self._table_names

    @property
    def limit(self):
        return self._limit

    def is_select(self):
        return self.parsed[0].get_type() == 'SELECT'

    def get_main_parsed_type(self):
        return self.parsed[0].get_type()

    def is_explain(self):
        return self.stripped().upper().startswith('EXPLAIN')

    def is_readonly(self):
        return self.is_select() or self.is_explain()

    def stripped(self):
        return self.sql.strip(' \t\n;')

    def get_statements(self):
        statements = []
        for statement in self.parsed:
            if statement:
                sql = str(statement).strip(' \n;\t')
                if sql:
                    statements.append(sql)
        return statements

    @staticmethod
    def __precedes_table_name(token_value):
        for keyword in PRECEDES_TABLE_NAME:
            if keyword in token_value:
                return True
        return False

    @staticmethod
    def get_full_name(identifier):
        if len(identifier.tokens) > 1 and identifier.tokens[1].value == '.':
            return '{}.{}'.format(identifier.tokens[0].value,
                                  identifier.tokens[2].value)
        return identifier.get_real_name()

    @staticmethod
    def __is_result_operation(keyword):
        for operation in RESULT_OPERATIONS:
            if operation in keyword.upper():
                return True
        return False

    @staticmethod
    def __is_identifier(token):
        return isinstance(token, (IdentifierList, Identifier))

    def __process_identifier(self, identifier):
        if '(' not in '{}'.format(identifier):
            self._table_names.add(self.get_full_name(identifier))
            return

        # store aliases
        if hasattr(identifier, 'get_alias'):
            self._alias_names.add(identifier.get_alias())
        if hasattr(identifier, 'tokens'):
            # some aliases are not parsed properly
            if identifier.tokens[0].ttype == Name:
                self._alias_names.add(identifier.tokens[0].value)
        self.__extract_from_token(identifier)

    def as_create_table(self, table_name, overwrite=False):
        exec_sql = ''
        sql = self.stripped()
        if overwrite:
            exec_sql = 'DROP TABLE IF EXISTS {};\n'.format(table_name)
        exec_sql += 'CREATE TABLE {} AS \n{}'.format(table_name, sql)
        return exec_sql

    def __extract_from_token(self, token):
        if not hasattr(token, 'tokens'):
            return

        table_name_preceding_token = False

        for item in token.tokens:
            if item.is_group and not self.__is_identifier(item):
                self.__extract_from_token(item)

            if item.ttype in Keyword:
                if self.__precedes_table_name(item.value.upper()):
                    table_name_preceding_token = True
                    continue

            if not table_name_preceding_token:
                continue

            if item.ttype in Keyword or item.value == ',':
                if (self.__is_result_operation(item.value) or
                        item.value.upper() == ON_KEYWORD):
                    table_name_preceding_token = False
                    continue
                # FROM clause is over
                break

            if isinstance(item, Identifier):
                self.__process_identifier(item)

            if isinstance(item, IdentifierList):
                for token in item.tokens:
                    if self.__is_identifier(token):
                        self.__process_identifier(token)

    def _get_limit_from_token(self, token):
        if token.ttype == sqlparse.tokens.Literal.Number.Integer:
            return int(token.value)
        elif token.is_group:
            return int(token.get_token_at_offset(1).value)

    def _extract_limit_from_query(self, statement):
        limit_token = None
        for pos, item in enumerate(statement.tokens):
            if item.ttype in Keyword and item.value.lower() == 'limit':
                limit_token = statement.tokens[pos + 2]
                return self._get_limit_from_token(limit_token)

    def get_query_with_new_limit(self, new_limit):
        if not self._limit:
            return self.sql + ' LIMIT ' + str(new_limit)
        limit_pos = None
        tokens = self.parsed[0].tokens
        # Add all items to before_str until there is a limit
        for pos, item in enumerate(tokens):
            if item.ttype in Keyword and item.value.lower() == 'limit':
                limit_pos = pos
                break
        limit = tokens[limit_pos + 2]
        if limit.ttype == sqlparse.tokens.Literal.Number.Integer:
            tokens[limit_pos + 2].value = new_limit
        elif limit.is_group:
            tokens[limit_pos + 2].value = (
                '{}, {}'.format(next(limit.get_identifiers()), new_limit)
            )

        str_res = ''
        for i in tokens:
            str_res += str(i.value)
        return str_res

    def get_v(self, tok):
        for token in tok.tokens:
            if token._get_repr_name() in 'Parenthesis':
                # print("ParenthesisParenthesisParenthesisParenthesis", token)
                self.get_v(token)
            elif token._get_repr_name() == 'Values':
                # print("ValuesValuesValuesValues", token)
                self.get_v(token)
            elif token._get_repr_name() == 'IdentifierList':
                # print("IdentifierListIdentifierListIdentifierListIdentifierList", token)
                self.get_v(token)
            elif token._get_repr_name() in ('Punctuation', 'Whitespace', 'Newline'):
                pass
            elif token._get_repr_name() in ('Keyword',) and token.value != 'NULL':
                pass
            elif token._get_repr_name() == 'DML':
                self.sql_type = token._get_repr_value()
            elif token._get_repr_name() == 'Identifier':
                self.get_v(token)
            elif token._get_repr_name() == 'Name':
                if token.value.startswith('`'):
                    if self.table_name:
                        # self.keys.append({'type': token._get_repr_name(), 'value': token.value.replace('`', ''), 'last': tok._get_repr_name()})
                        self.keys.append(token.value.replace('`', ''))
                    else:
                        self.table_name = token.value.replace('`', '')
            elif token._get_repr_name() == 'Function':
                self.get_v(token)
            else:
                self.values.append({'type': token._get_repr_name(), 'value': token.value})
                # pprint_dir(tokens)

    def get_v_v2(self, tok):
        for token in tok.flatten():
            if token._get_repr_name() in ('Parenthesis', 'Values', 'IdentifierList'):
                pass
            elif token._get_repr_name() in ('Punctuation', 'Whitespace', 'Newline'):
                pass
            elif token._get_repr_name() in ('Keyword',) and token.value != 'NULL':
                pass
            elif token._get_repr_name() == 'DML':
                self.sql_type = token._get_repr_value()
            elif token._get_repr_name() == 'Identifier':
                pass
            elif token._get_repr_name() == 'Name':
                if token.value.startswith('`'):
                    if self.table_name:
                        # self.keys.append({'type': token._get_repr_name(), 'value': token.value.replace('`', ''), 'last': tok._get_repr_name()})
                        self.keys.append(token.value.replace('`', ''))
                    else:
                        self.table_name = token.value.replace('`', '')
            elif token._get_repr_name() == 'Function':
                pass
            else:
                self.values.append({'type': token._get_repr_name(), 'value': token.value})


class SqlExtractor(BaseExtractor):
    """提取sql语句"""

    @staticmethod
    def get_full_name(identifier, including_dbs=False):
        if len(identifier.tokens) > 1 and identifier.tokens[1].value == '.':
            a = identifier.tokens[0].value
            b = identifier.tokens[2].value
            db_table = (a, b)
            full_tree = '{}.{}'.format(a, b)
            if len(identifier.tokens) == 3:
                return full_tree
            else:
                i = identifier.tokens[3].value
                c = identifier.tokens[4].value
                if i == ' ':
                    return full_tree
                full_tree = '{}.{}.{}'.format(a, b, c)
                return full_tree
        return None, None

    sql_type = None
    keys = []
    values = []
    data = []
    table_name = None


import sys
import os
import io


class Logger(object):
    def __init__(self):
        self.terminal = sys.stdout
        self.log = io.StringIO()
        self.path = os.path.abspath(os.path.dirname(__file__))
        self.type = sys.getfilesystemencoding()

    def write(self, message):
        # self.terminal.write(message)
        # print(message)
        self.log.write(message)
        self.log.readable()

    def flush(self):
        self.log.flush()


def pprint_dir(data, get_attr='', skip='__', print_func=True, run_func=True, print_attr=True):
    """
    获取一个参数的所有属性 以及方法 的值 打印出来

    get_attr: 尝试获取单个属性
    """

    print(f'1111111111111111111111111 pprint_dir 1111111111111111111111111        '
          f'[ type: {str(type(data))[8:-2]:25s} ]      [ __repr__: {data.__repr__()}] {data.__str__().__repr__()}')
    for attr in dir(data):
        if skip and attr.startswith(skip):
            continue
        if get_attr and get_attr != attr:
            continue
        d_attr = getattr(data, attr)

        if callable(d_attr):
            if not print_func:
                continue
            if run_func:
                old = sys.stdout
                try:
                    tmp = Logger()
                    sys.stdout = tmp
                    res = d_attr()
                    sys.stdout = old
                    value = f'[   res][ rtype: {str(type(res))[8:-2]:65s}] {res}   |  {tmp.log.getvalue()}'
                    # value = f'[funcR][{str(type(res))[8:-2]:25}] {res}'
                except Exception as ex:
                    value = f'[except][ etype: {str(type(ex))[8:-2]:65s}] {ex}'
                finally:
                    sys.stdout = old
            else:
                value = '[  func][  fdoc: {:65s}] {}'.format((d_attr.__doc__ or '').split("\n")[0], d_attr)
        else:
            if not print_attr:
                continue
            value = f'[  attr][ atype: {str(type(d_attr))[8:-2]:65s}] {d_attr}'
        print('[{:25s}]{}'.format(attr, value.replace("\n", "\\n")))

    print(f'2222222222222222222222222 pprint_dir 2222222222222222222222222        '
          f'[ type: {str(type(data))[8:-2]:25s} ]      [ __repr__: {data.__repr__()}] {data.__str__().__repr__()}\n')


def main111():
    with open('/home/user/Desktop/test.sql', 'r', encoding='utf-8') as f:
        table_info_str = ''
        flag_create_table = False
        table_keys = []

        def is_insert(li):
            return li.lower()[:7] in ('insert ', 'replace')

        def get_table_keys(table_info_str_raw):
            if table_info_str_raw.endswith(';'):
                table_info_str_raw = table_info_str_raw[-1]
            info_str = f'{table_info_str_raw});'
            return list(SqlExtractor(info_str).get_v_main()[0].keys())

        for line in f:
            line = line.strip()
            if line.lower().startswith('create table'):
                flag_create_table = True
            elif line.lower() == '-- ----------------------------':
                flag_create_table = False
                if table_info_str and not table_keys:
                    table_keys = get_table_keys(table_info_str)
            elif is_insert(line):
                flag_create_table = False
                if table_info_str and not table_keys:
                    table_keys = get_table_keys(table_info_str)
            if flag_create_table:
                if line.strip().startswith('`') or line.strip().lower().startswith('create table'):
                    table_info_str += line
            if is_insert(line):
                yield from match_insert(line)
                # for data in SqlExtractor(line.strip()).get_v_main(table_keys):
                #     yield data
                # 在正则表达式最前面加上 (?i) 就可以忽略后面所有的大小写


RANDOM_STR = '~!}#:~!|F"SF:}~!@WFR{:FE!@#R#3p[r23kef[1p2343213vdgbv21331r123rfefe2g23geb|}}{}{@!{}djk(#JFD)P!#!LSFD"21r1rff43>F{:DS'


def format_col(col):
    if len(col) == 0:
        return ''
    elif col == 'NULL':
        return None
    elif isinstance(col, str):
        col = col.replace("\\'", "'").replace(RANDOM_STR, "\\")
        if col.startswith("'") and col.endswith("'"):
            col = col[1:-1]
        elif not re.sub('[0-9]', '', col):
            col = int(col)
        return col
    else:
        return col


def parse_values(values):
    """
    Given a file handle and the raw values from a MySQL INSERT
    statement, write the equivalent CSV to the file
    """

    delimiter = ', '
    quotechar = "'"
    latest_row = []
    reader = csv.reader(
        [values.replace("\\\\", RANDOM_STR).replace("\\", "\\\\").replace('\\"', f'{RANDOM_STR}"')],
        # v:= [values.replace("\\'", "\\\\'")],
        # [values.replace("\\'", random_str)],
        # [values],
        delimiter=',',
        doublequote=False,
        # skipinitialspace=True,
        escapechar='\\',
        quotechar=quotechar,
        strict=True
    )
    # reader = re.split(f"{delimiter}(?=([^{quotechar}]*{quotechar}[^{quotechar}]*{quotechar})*[^{quotechar}]*$)", values)
    # reader = re.findall(f'[^{delimiter}{quotechar}]+|{delimiter}{delimiter}|(?:{quotechar}[^{delimiter}{quotechar}]*{quotechar}[^{quotechar}]*{quotechar}[^{quotechar}]*){quotechar}|{quotechar}(?:[^{quotechar}])*{quotechar}', values)
    last_col = ''
    last_split = False
    # reader = re.split(f'(?:^|({delimiter}))(?:{quotechar}(([^{quotechar}]*)+((?:{quotechar}{quotechar}([^{quotechar}]*)+)*)+){quotechar}|(((^((?!{delimiter}|{quotechar}).)+$)*)+))', values)
    # for i in range(1):
    #     for column in reader:

    for reader_row in reader:
        for idx, column in enumerate(reader_row):
            if column == f"{delimiter}{quotechar}":
                last_col += ",'"
                latest_row.append(format_col(last_col))
                latest_row.append('')
                last_col = ""
                column = ""
            if column.startswith(f"{delimiter}"):
                last_col += ",'"
                latest_row.append(format_col(last_col))
                last_col = "'"
                last_split = True
                column = column[2:]
            if len(column) and column[0] == ' ' and not last_col:
                column = column[1:]
            if not last_col and (not column or delimiter == column):
                continue
            # column = column.strip()
            # If our current string is empty...
            if len(column) == 0 and not last_col:
                latest_row.append('')
                continue
            elif column == 'NULL':
                latest_row.append(None)
                continue
            # If our string starts with an open paren
            if column and column[0] == "(":
                # Assume that this column does not begin
                # a new row.
                new_row = False
                # If we've been filling out a row
                if len(latest_row) > 0:
                    # Check if the previous entry ended in
                    # a close paren. If so, the row we've
                    # been filling out has been COMPLETED
                    # as:
                    #    1) the previous entry ended in a )
                    #    2) the current entry starts with a (
                    # print(latest_row)
                    if latest_row[-1][-1] == ")":
                        # Remove the close paren.
                        latest_row[-1] = format_col(latest_row[-1][:-1])
                        new_row = True
                # If we've found a new row, write it out
                # and begin our new one
                if new_row:
                    yield latest_row
                    latest_row = []
                # If we're beginning a new row, eliminate the
                # opening parentheses.
                if len(latest_row) == 0:
                    column = column[1:]
            # Add our column to the row we're working on.

            now_col = format_col(column)
            if (
                    (not last_col) and
                    column.startswith("'") and
                    (
                            (
                                    (
                                            not (
                                                    (column[-1:] == "'" and column[-2:-1] != '\\') or
                                                    (column.endswith("')") and column[-3:-2] != '\\')
                                            )
                                    )
                            ) or
                            len(column) == 1
                    )
            ):
                last_col = column
            elif last_col:
                if column != delimiter:
                    if last_split:
                        last_col += f'{column}'
                    else:
                        last_col += f',{column}'
                    last_split = False
                else:
                    last_col += delimiter
                    last_split = True
                if column[-1:] == "'" and column[-2:-1] != '\\':
                    latest_row.append(format_col(last_col))
                    last_col = ''
                elif column.endswith("')") and column[-3:-2] != '\\':
                    latest_row.append(last_col)
                    last_col = ''
            else:
                latest_row.append(now_col)
        # At the end of an INSERT statement, we'll
        # have the semicolon.
        # Make sure to remove the semicolon and
        # the close paren.
        if latest_row and latest_row[-1]:
            # if latest_row[-1][-2:] == ");":
            #     latest_row[-1] = format_col(latest_row[-1][:-2])
            #     yield latest_row
            if latest_row[-1][-1:] == ")":
                latest_row[-1] = format_col(latest_row[-1][:-1])
                yield latest_row


def match_insert(line):
    m = re.fullmatch(
        f'(?i)(INSERT|REPLACE|INSERT IGNORE) INTO\s?`?(?P<table_name>.+?)`?\s?(\((?P<keys>.+?)\)\s)?VALUES\s?(?P<values>(\(.+\)[\s,]?)+);?',
        line)
    table_name, keys, values = '', [], []
    if m:
        table_name = m.group('table_name')
        keys = [re.sub(r'[`\'\"\s]', '', key) for key in m.group('keys').strip().split(',')] if m.group('keys') else []
        # print([key for key in re.split(r'\),\s?\(', m.group('values').strip())])
        # print(m.group('values').strip())
        values = [_ for _ in parse_values(m.group('values').strip())]

    return table_name, keys, values


if __name__ == '__main__':

    # for i in parse_values(r"""(503802367, NULL, '(｡•ˇ‸ˇ•｡) 为中华之崛起添砖加瓦', '❤️', 'lovefeng')"""):
    # for i in parse_values(r"""(1418913438, NULL, 'faku', 'gh6g,, ,,,', '')"""):
    # for i in parse_values(r"""(1023904092, NULL, 'JIANG,', 'Ben', '')"""):
    for i in parse_values(r"""(2, 1, NULL, b'0000000000000000000000000000000000000000000000000000000000000001', NULL, '33', '2020-12-09', '2020-12-09 17:52:44.000000', 234, 2134, '1222', 123, NULL, NULL, 234, 13, '{\"1\": \"2\"}', NULL, NULL, '234', NULL, 234, '123r', NULL, NULL, NULL, 1234, NULL, NULL, 1234, '', 1234, '1234', '17:57:22.000000', '2020-12-09 17:57:25.000000', NULL, 1, 'd213d', NULL, '23r2e132r1')"""):
    # for i in parse_values(r"""(1023904092, NULL, 'JIANG,'', 'Ben', '')"""):
    # for i in parse_values(r"""(952012376, NULL, 'Remy', '\" 孙琪\"', '')"""):
        print(i)

    # sql = """select
    # b.product_name "产品",
    # count(a.order_id) "订单量",
    # b.selling_price_max "销售价",
    # b.gross_profit_rate_max/100 "毛利率",
    # case when b.business_type =1 then '自营消化' when b.business_type =2 then '服务商消化'  end "消化模式"
    # from(select 'CRM签单' label,date(d.update_ymd) close_ymd,c.product_name,c.product_id,
    #     a.order_id,cast(a.recipient_amount as double) amt,d.cost
    #     from mysql4.dataview_fenxiao.fx_order a
    #     left join mysql4.dataview_fenxiao.fx_order_task b on a.order_id = b.order_id
    #     left join mysql7.dataview_trade.ddc_product_info c on cast(c.product_id as varchar) = a.product_ids and c.snapshot_version = 'SELLING'
    #     inner join (select t1.par_order_id,max(t1.update_ymd) update_ymd,
    #                 sum(case when t4.product2_type = 1 and t5.shop_id is not null then t5.price else t1.order_hosted_price end) cost
    #                from hive.bdc_dwd.dw_mk_order t1
    #                left join hive.bdc_dwd.dw_mk_order_status t2 on t1.order_id = t2.order_id and t2.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #                left join mysql7.dataview_trade.mk_order_merchant t3 on t1.order_id = t3.order_id
    #                left join mysql7.dataview_trade.ddc_product_info t4 on t4.product_id = t3.MERCHANT_ID and t4.snapshot_version = 'SELLING'
    #                left join mysql4.dataview_scrm.sc_tprc_product_info t5 on t5.product_id = t4.product_id and t5.shop_id = t1.seller_id
    #                where t1.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #                and t2.valid_state in (100,200) ------有效订单
    #                and t1.order_mode = 10    --------产品消耗订单
    #                and t2.complete_state = 1  -----订单已经完成
    #                group by t1.par_order_id
    #     ) d on d.par_order_id  = b.task_order_id
    #     where c.product_type = 0 and date(from_unixtime(a.last_recipient_time)) > date('2016-01-01') and a.payee_type <> 1 -----------已收款
    #     UNION ALL
    #     select '企业管家消耗' label,date(c.update_ymd) close_ymd,b.product_name,b.product_id,
    #     a.task_id,(case when a.yb_price = 0 and b.product2_type = 1 then b.selling_price_min else a.yb_price end) amt,
    #     (case when a.yb_price = 0 and b.product2_type = 2 then 0 when b.product2_type = 1 and e.shop_id is not null then e.price else c.order_hosted_price end) cost
    #     from mysql8.dataview_tprc.tprc_task a
    #     left join mysql7.dataview_trade.ddc_product_info b on a.product_id = b.product_id and b.snapshot_version = 'SELLING'
    #     inner join hive.bdc_dwd.dw_mk_order c on a.order_id = c.order_id and c.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #     left join hive.bdc_dwd.dw_mk_order_status d on d.order_id = c.order_id and d.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #     left join mysql4.dataview_scrm.sc_tprc_product_info e on e.product_id = b.product_id and e.shop_id = c.seller_id
    #     where  d.valid_state in (100,200) and d.complete_state = 1  and c.order_mode = 10
    #     union ALL
    #     select '交易管理系统' label,date(t6.close_ymd) close_ymd,t4.product_name,t4.product_id,
    #     t1.order_id,(t1.order_hosted_price-t1.order_refund_price) amt,
    #     (case when t1.order_mode <> 11 then t7.user_amount when t1.order_mode = 11 and t4.product2_type = 1 and t5.shop_id is not null then t5.price else t8.cost end) cost
    #     from hive.bdc_dwd.dw_mk_order t1
    #     left join hive.bdc_dwd.dw_mk_order_business t2 on t1.order_id = t2.order_id and t2.acct_day=substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #     left join mysql7.dataview_trade.mk_order_merchant t3 on t1.order_id = t3.order_id
    #     left join mysql7.dataview_trade.ddc_product_info t4 on t4.product_id = t3.MERCHANT_ID and t4.snapshot_version = 'SELLING'
    #     left join mysql4.dataview_scrm.sc_tprc_product_info t5 on t5.product_id = t4.product_id and t5.shop_id = t1.seller_id
    #     left join hive.bdc_dwd.dw_fact_task_ss_daily t6 on t6.task_id = t2.task_id and t6.acct_time=date_format(date_add('day',-1,current_date),'%Y-%m-%d')
    #     left join (select a.task_id,sum(a.user_amount) user_amount
    #                from hive.bdc_dwd.dw_fn_deal_asyn_order a
    #                where a.is_new=1 and a.service='Trade_Payment' and a.state=1 and a.acct_day=substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #                group by a.task_id)t7 on t7.task_id = t2.task_id
    #     left join (select t1.par_order_id,sum(t1.order_hosted_price - t1.order_refund_price) cost
    #                from hive.bdc_dwd.dw_mk_order t1
    #                where t1.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2) and t1.order_type = 1 and t1.order_stype = 4 and t1.order_mode = 12
    #                group by t1.par_order_id) t8 on t1.order_id = t8.par_order_id
    #     where t1.acct_day = substring(cast(DATE_ADD('day',-1,CURRENT_DATE) as varchar),9,2)
    #     and t1.order_type = 1 and t1.order_stype in (4,5) and t1.order_mode <> 12 and t4.product_id is not null and t1.order_hosted_price > 0 and t6.is_deal = 1 and t6.close_ymd >= '2018-12-31'
    # )a
    # left join mysql7.dataview_trade.ddc_product_info b on a.product_id = b.product_id and b.snapshot_version = 'SELLING'
    # where b.product2_type = 1 -------标品
    # and close_ymd between DATE_ADD('day',-7,CURRENT_DATE)  and DATE_ADD('day',-1,CURRENT_DATE)
    # GROUP BY b.product_name,
    # b.selling_price_max,
    # b.gross_profit_rate_max/100,
    # b.actrul_supply_num,
    # case when b.business_type =1 then '自营消化' when b.business_type =2 then '服务商消化'  end
    # order by count(a.order_id) desc
    # limit 10"""
    #
    # sql = '''
    # INSERT INTO `chat_media`(`fromid`, `toidfix`, `msgid`, `toid`, `dialogid`, `clientid`, `chatid`, `totype`, `msgtype`, `size`, `realsize`, `timeupdate`, `dateint`, `date`, `grouped_id`, `md5`, `url`, `display_url`, `status`, `message`, `obj`, `filename`, `savepath`, `error`, `location`, `timeupload`) VALUES (0, -1001401435758, 2466, 1401435758, '1401435758', 1250199136, 1401435758, 'PeerChannel', 'PhotoVideo', 92371, 78702, 1605035384, 1589634510, '2020-05-16 13:08:30.000000', NULL, NULL, NULL, NULL, 3, NULL, 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one.jpg', 'PhotoVideo/1401435758/2466--2020-05-16T21_08_30+08_00--None.jpg', NULL, NULL, 1606712817);
    # '''
    sql = '''
    /*
 Navicat Premium Data Transfer

 Source Server         : local
 Source Server Type    : MySQL
 Source Server Version : 50727
 Source Host           : localhost:3306
 Source Schema         : 122_

 Target Server Type    : MySQL
 Target Server Version : 50727
 File Encoding         : 65001

 Date: 09/12/2020 17:58:13
*/

SET NAMES utf8mb4;
SET FOREIGN_KEY_CHECKS = 0;

-- ----------------------------
-- Table structure for test
-- ----------------------------
DROP TABLE IF EXISTS `test`;
CREATE TABLE `test`  (
  `id` bigint(64) NOT NULL AUTO_INCREMENT,
  `bigint` bigint(255) NULL DEFAULT NULL,
  `binary` binary(255) NULL DEFAULT NULL,
  `bit` bit(64) NULL DEFAULT NULL,
  `blob` blob NULL,
  `char` char(255) CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL DEFAULT NULL,
  `date` date NULL DEFAULT NULL,
  `datetime` datetime(6) NULL DEFAULT NULL,
  `decimal` decimal(65, 0) NULL DEFAULT NULL,
  `double` double(255, 0) NULL DEFAULT NULL,
  `enum` enum('123','1222') CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL DEFAULT NULL,
  `float` float(255, 0) NULL DEFAULT NULL,
  `geometry` geometry NULL,
  `geometrycollection` geometrycollection NULL,
  `int` int(255) NULL DEFAULT NULL,
  `integer` int(255) NULL DEFAULT NULL,
  `json` json NULL,
  `linestring` linestring NULL,
  `longblob` longblob NULL,
  `longtext` longtext CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL,
  `mediumblob` mediumblob NULL,
  `mediumint` mediumint(255) NULL DEFAULT NULL,
  `mediumtext` mediumtext CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL,
  `multilinestring` multilinestring NULL,
  `multipoint` multipoint NULL,
  `multipolygon` multipolygon NULL,
  `numeric` decimal(65, 0) NULL DEFAULT NULL,
  `point` point NULL,
  `polygon` polygon NULL,
  `real` double(255, 0) NULL DEFAULT NULL,
  `set` set('') CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL DEFAULT '',
  `smallint` smallint(255) NULL DEFAULT NULL,
  `text` text CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL,
  `time` time(6) NULL DEFAULT NULL,
  `timestamp` timestamp(6) NULL DEFAULT NULL,
  `tinyblob` tinyblob NULL,
  `tinyint` tinyint(255) NULL DEFAULT NULL,
  `tinytext` tinytext CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL,
  `varbinary` varbinary(255) NULL DEFAULT NULL,
  `varchar` varchar(255) CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_520_ci NULL DEFAULT NULL,
  PRIMARY KEY (`id`) USING BTREE
) ENGINE = InnoDB AUTO_INCREMENT = 3 CHARACTER SET = utf8mb4 COLLATE = utf8mb4_unicode_520_ci ROW_FORMAT = Dynamic;

-- ----------------------------
-- Records of test
-- ----------------------------
INSERT INTO `test` VALUES (1, 1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, '', NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
INSERT INTO `test` VALUES (2, 1, NULL, b'0000000000000000000000000000000000000000000000000000000000000001', NULL, '33', '2020-12-09', '2020-12-09 17:52:44.000000', 234, 2341, '1222', 23, NULL, NULL, 234, 13, '{\"1\": \"2\"}', NULL, NULL, '234', NULL, 234, '123r', NULL, NULL, NULL, 1234, NULL, NULL, 1234, '', 1234, '1234', '17:57:22.000000', '2020-12-09 17:57:25.000000', NULL, 1, 'd213d', NULL, '23r2e132r1'),(2, 1, NULL, b'0000000000000000000000000000000000000000000000000000000000000001', NULL, '33', '2020-12-09', '2020-12-09 17:52:44.000000', 234, 2341, '1222', 23, NULL, NULL, 234, 13, '{\"1\": \"2\"}', NULL, NULL, '234', NULL, 234, '123r', NULL, NULL, NULL, 1234, NULL, NULL, 1234, '', 1234, '1234', '17:57:22.000000', '2020-12-09 17:57:25.000000', NULL, 1, 'd213d', NULL, '6566666(2, 1, NULL, b'0000000000000000000000000000000000000000000000000000000000000001', NULL, '33', '2020-12-09', '2020-12-09 17:52:44.000000', 234, 2341, '1222', 23, NULL, NULL, 234, 13, '{\"1\": \"2\"}', NULL, NULL, '234', NULL, 234, '123r', NULL, NULL, NULL, 1234, NULL, NULL, 1234, '', 1234, '1234', '17:57:22.000000', '2020-12-09 17:57:25.000000', NULL, 1, 'd213d', NULL, '23r2e132r1'),(2, 1, NULL, b'0000000000000000000000000000000000000000000000000000000000000001', NULL, '33', '2020-12-09', '2020-12-09 17:52:44.000000', 234, 2341, '1222', 23, NULL, NULL, 234, 13, '{\"1\": \"2\"}', NULL, NULL, '234', NULL, 234, '123r', NULL, NULL, NULL, 1234, NULL, NULL, 1234, '', 1234, '1234', '17:57:22.000000', '2020-12-09 17:57:25.000000', NULL, 1, 'd213d', NULL, '434311111111111111111111111111112')');

SET FOREIGN_KEY_CHECKS = 1;
    '''
    # sql_extractor = SqlExtractor(sql)
    import random

    # print(sql_extractor.sql)

    # sql_extractor.get_v(sxql_extractor.parsed[0].tokens)



import diskcache
import json
import os
import zlib
import time
import sys
import functools
import asyncio

PATH_REL_DATA = 'data_disk_cache'

PATH_CWD = os.getcwd()
path_abs_base_add = functools.partial(os.path.join, PATH_CWD)
PATH_ABS_DATA = path_abs_base_add(PATH_REL_DATA)
path_abs_data_add = functools.partial(os.path.join, PATH_ABS_DATA)


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]


sys.path.append(os.path.join(get_realpath(), './'))


def str_to_int_time(string):
    return time.mktime(time.strptime(string, '%Y-%m-%d %H:%M:%S.%f')) - 3600 * 8


def int_to_str_time_gmtime(timeint):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(timeint))


def int_to_str_time_localtime(timeint):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(timeint))


def int_to_str_time_second(timeint):
    return time.strftime('%H:%M:%S', time.gmtime(timeint / 1000))


def get_time_day():
    return time.strftime("%Y-%m-%d", time.localtime())


def get_time_day_all():
    return time.strftime("%Y_%m_%d__%H_%M_%S", time.localtime())


def get_make_local_path(*path):
    local_path = os.path.join(*path)
    if not os.path.exists(local_path):
        os.makedirs(local_path)
    return local_path


class JSONDisk(diskcache.Disk):
    def __init__(self, directory, compress_level=1, **kwargs):
        self.compress_level = compress_level
        super(JSONDisk, self).__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = json.dumps(key).encode('utf-8')
        data = zlib.compress(json_bytes, self.compress_level)
        return super(JSONDisk, self).put(data)

    def get(self, key, raw):
        data = super(JSONDisk, self).get(key, raw)
        return json.loads(zlib.decompress(data).decode('utf-8'))

    def store(self, value, read, key=None):
        if not read:
            json_bytes = json.dumps(value).encode('utf-8')
            value = zlib.compress(json_bytes, self.compress_level)
        return super(JSONDisk, self).store(value, read)

    def fetch(self, mode, filename, value, read):
        data = super(JSONDisk, self).fetch(mode, filename, value, read)
        if not read:
            data = json.loads(zlib.decompress(data).decode('utf-8'))
        return data


class JSONDiskNoC(diskcache.Disk):
    def __init__(self, directory, compress_level=0, **kwargs):
        self.compress_level = compress_level
        super(JSONDiskNoC, self).__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = json.dumps(key).encode('utf-8')
        return super(JSONDiskNoC, self).put(json_bytes)

    def get(self, key, raw):
        data = super(JSONDiskNoC, self).get(key, raw)
        return json.loads(data.decode('utf-8'))

    def store(self, value, read, key=None):
        if not read:
            value = json.dumps(value).encode('utf-8')
        return super(JSONDiskNoC, self).store(value, read)

    def fetch(self, mode, filename, value, read):
        data = super(JSONDiskNoC, self).fetch(mode, filename, value, read)
        if not read:
            data = json.loads(data.decode('utf-8'))
        return data


cache_default = None


def init_cache_default():
    cache_default = diskcache.Cache(
        directory=PATH_ABS_DATA, disk=JSONDiskNoC, disk_compress_level=6,
        size_limit=int(1024 ** 4), cull_limit=0)
    cache_default.create_tag_index()


def get_cache_default(*args, **kwargs):
    return cache_default.get(*args, **kwargs)


def set_cache_default(*args, **kwargs):
    with cache_default.transact():
        return cache_default.set(*args, **kwargs)


async def set_cache_default_async(key, value, expire=None, read=False, tag=None, retry=False):
    loop = asyncio.get_running_loop()
    future = loop.run_in_executor(None, cache_default.set, key, value, expire, read, tag, retry)
    result = await future
    return result


def list_cache_default(keys=None):
    for k in cache_default.iterkeys():
        if keys and k not in keys:
            continue
        else:
            print(f'[cache_print] {k}: {cache_default[k]}')


def cull_cache_default():
    """
    将缓存目录清理到只剩size_limit的大小
    """
    return cache_default.cull(retry=True)


# print(cache_default.size_limit, cache_default.disk_min_file_size, cache_default.cull_limit)
# cache.reset('cull_limit', 0)  # Disable automatic evictions.
if __name__ == '__main__':
    set_cache_default('rwfrwr', 1)
    print(PATH_ABS_DATA, 'tag_index开启' if cache_default.tag_index else 'tag_index尚未开启，请执行,cache_default.create_tag_index()', f'[{cache_default.volume()/1024/1024/1024:.8f}/{cache_default.size_limit/1024/1024/1024:.8f}GB]', cache_default.disk_min_file_size, cache_default.cull_limit)
    list_cache_default()
    # with cache_default.transact():
    #     # cache.set('test', 111)
    #     print(cache_default.get('test'))
        # asyncio.run(set_async('test-key', 'test-value'))



import copy
import sys

import nb_log
import pymysql
import paramiko
import typing
from concurrent.futures import ThreadPoolExecutor
from threadpool_executor_shrink_able import BoundedThreadPoolExecutor
from universal_object_pool import ObjectPool, AbstractObject
import time
import decorator_libs
from itertools import chain

from d22d.utils.decorators import print_hz


class TimeoutTransport(paramiko.Transport):
    _active_check_timeout = 100


t = TimeoutTransport(('127.0.0.1', 22))



class PyMysqlOperator(AbstractObject):
    error_type_list_set_not_available = []  # 出了特定类型的错误，可以设置对象已经无效不可用了，不归还到队列里面。

    # error_type_list_set_not_available = [pymysql.err.InterfaceError]

    def __init__(self, host='192.168.6.130', user='root', password='123456', cursorclass=pymysql.cursors.SSDictCursor,
                 autocommit=False, **pymysql_connection_kwargs):
        super().__init__()
        in_params = copy.copy(locals())
        in_params.update(pymysql_connection_kwargs)
        in_params.pop('self')
        in_params.pop('pymysql_connection_kwargs')
        in_params.pop('__class__')
        self.in_params = in_params
        self.is_recon = False
        self.conn = None
        self.reconnect(in_params)
        self.logger = nb_log.get_logger(self.__class__.__name__)

    def reconnect(self, in_params):
        while True:
            try:
                self.conn = pymysql.Connection(**in_params)
                if self.is_recon:
                    print('重新连接成功。。。')
                break
            except pymysql.err.OperationalError as ex:
                if ex.args == (1040, 'ny connections'):
                    print('mysql连接数过多，15秒后自动重试重新连接。。。')
                    time.sleep(15)
                    self.is_recon = True
                else:
                    raise
    """ 下面3个是重写的方法"""

    def clean_up(self):  # 如果一个对象最近30分钟内没被使用，那么对象池会自动将对象摧毁并从池中删除，会自动调用对象的clean_up方法。
        self.conn.close()

    def before_use(self):
        self.cursor = self.conn.cursor()
        self.core_obj = self.cursor  # 这个是为了operator对象自动拥有cursor对象的所有方法。

    def before_back_to_queue(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.conn.rollback()
        else:
            self.conn.commit()
        self.cursor.close()  # 也可以不要，因为每次的cusor都是不一样的。

    """以下可以自定义其他方法。
    因为设置了self.core_obj = self.cursor ，父类重写了__getattr__,所以此对象自动拥有cursor对象的所有方法,如果是同名同意义的方法不需要一个个重写。
    """

    def execute(self, query, args=None):
        """
        这个execute由于方法名和入参和逻辑与官方一模一样，可以不需要，因为设置了core_obj后，operator对象自动拥有cursor对象的所有方法，可以把这个方法注释了然后测试运行不受影响。
        :param query:
        :param args:
        :return:
        """
        return self.cursor.execute(query, args)


if __name__ == "__main__":
    # local_ctr_mysql = MYSQLController(host='127.0.0.1', port=3306, user='mq', passwd='1234qwer', database='gu')
    # local_execute_auto = local_ctr_mysql.execute_auto
    # print(local_execute_auto("select count(0) as cnt from gu1", return_type=1))  # 第一个字典
    # print(local_execute_auto("select count(0) as cnt from gu1", return_type=2))  # 字典列表
    # print(local_execute_auto("select count(0) as cnt from gu1", return_type=3))  # 字典迭代器
    # print(local_execute_auto("select count(0) as cnt from gu1", return_type=4))  # 第一列第一行的值

    mysql_pool = ObjectPool(object_type=PyMysqlOperator, object_pool_size=1000, object_init_kwargs=dict(
        host='127.0.0.1', port=3306, user='mq', password='1234qwer'))

    iii = 0
    time_l = time.time()


    # @print_hz('test_update', per_cnt_print=1000000, print_func=print)
    # def test_update(i):
    #     return i

    @print_hz('test_update', print_func=print)
    def tes_update(i):
        sql = f'''
                INSERT INTO gu.testu(uname , age , numfff)
            VALUES{','.join(['(%s, %s, %s)'] * 1000)}

            ON DUPLICATE KEY UPDATE
                uname = values(uname),
                age = if(values(age)>age,values(age),age),
                numfff = if(values(numfff)>numfff,values(numfff),numfff);
            '''
        with mysql_pool.get(
                timeout=2) as operator:  # type: typing.Union[PyMysqlOperator,pymysql.cursors.DictCursor] #利于补全
            # print(id(operator.cursor), id(operator.conn))
            aa = ((f'name_{i+ii}', ((i+ii) * 4), 3245) for ii in range(1000))
            try:
                operator.execute(sql, args=list(chain.from_iterable(aa)))
            except pymysql.err.OperationalError as e:
                print(type(e),e.__dict__)
            # print(operator.lastrowid)  # opererator 自动拥有 operator.cursor 的所有方法和属性。 opererator.methodxxx 会自动调用 opererator.cursor.methodxxx


    # operator_global = PyMysqlOperator()
    #
    #
    # def test_update_multi_threads_use_one_conn(i):
    #     """
    #     这个是个错误的例子，多线程运行此函数会疯狂报错,单线程不报错
    #     这个如果运行在多线程同时操作同一个conn，就会疯狂报错。所以要么狠low的使用临时频繁在函数内部每次创建和摧毁mysql连接，要么使用连接池。
    #     :param i:
    #     :return:
    #     """
    #     sql = f'''
    #             INSERT INTO db1.table1(uname ,age)
    #         VALUES(
    #             %s ,
    #             %s)
    #         ON DUPLICATE KEY UPDATE
    #             uname = values(uname),
    #             age = if(values(age)>age,values(age),age);
    #         '''
    #
    #     operator_global.before_use()
    #     print(id(operator_global.cursor), id(operator_global.conn))
    #     operator_global.execute(sql, args=(f'name_{i}', i * 3))
    #     operator_global.cursor.close()
    #     operator_global.conn.commit()

    st = time.time()

    # 单线程 平均/峰值 [21.11/24.49] 次/秒
    # for x in range(0, 100):
    #     test_update(x)
    # print(time.time() - st)

    # pool_size 100 32多线程 平均/峰值 [343.59/377.96] 次/秒
    # pool_size 100 100多线程 平均/峰值 [995.77/1067.66] 次/秒
    #  pool_size 1000 220多线程 平均/峰值 [1821.46/1961.02] 次/秒
    #  pool_size 1000 280多线程 平均/峰值 [1732.67/2120.29] 次/秒
    #  pool_size 1000 320多线程 平均/峰值 [2391.82/2617.14] 次/秒
    #  pool_size 1000 320多线程 平均/峰值 [2130.53/2165.50] 次/秒
    #  pool_size 1000 320多线程 批量写1000 2个字段 平均/峰值 [15.64/15.80] 次/秒
    #  pool_size 1000 320多线程 批量写1000 3个字段 平均/峰值 [15.64/15.80] 次/秒
    #  pool_size 1000 400多线程 平均/峰值 [2204.73/2463.14] 次/秒
    #  pool_size 1000 640多线程 平均/峰值 [2231.81/2246.61] 次/秒
    # thread_pool = ThreadPoolExecutor(320)
    thread_pool = BoundedThreadPoolExecutor(320)
    with decorator_libs.TimerContextManager():
        for x in range(200000, 220000):
            thread_pool.submit(tes_update, x)
            # thread_pool.submit(test_update_multi_threads_use_one_conn, x)
    print(1231231111111111111111111111111111)
    #     thread_pool.shutdown()
    # time.sleep(10000)  # 这个可以测试验证，此对象池会自动摧毁连接如果闲置时间太长，会自动摧毁对象



#!/usr/bin/env python
# coding=utf-8
import codecs
import os
import sys
from version import __VERSION__

try:
    from setuptools import setup, find_packages
except:
    from distutils.core import setup

    find_packages = lambda x: []

# def read(fname):
#     return codecs.open(os.path.join(os.path.dirname(__file__), fname)).read()
#
#
# long_des = read("README.md")

with open('README.md', encoding='utf-8') as f:
    long_text = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    install_requires = [
        line for line in f.read().strip().splitlines()
        if not line.startswith('#')]

setup(
    name='d22d',
    version=__VERSION__,
    description=(
        'Migrating form DataBase to DataBase by 2 lines code'
    ),
    long_description=long_text,
    long_description_content_type="text/markdown",
    author='readerror',
    author_email='readerror@sina.com',
    maintainer='readerror',
    maintainer_email='readerror@sina.com',
    license='GPL License',
    packages=find_packages(),
    platforms=["all"],
    zip_safe=True,
    package_data={'': ['*']},
    url='https://github.com/DJMIN/D2D',
    python_requires='>=3.6',
    install_requires=install_requires,
    extras_require={
        'OracleD': [
            "cx_Oracle==7.2.3",
            # 'oracle数据库：11.2.0.2.0 - 64bit',
            # 'instantclient：11.2.0.4.0 - 64bit'
        ],
        'http': ["http-parser"],
        'rar': ["unrar-cffi"],
    }
)



# -*- coding: utf-8 -*-

from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer
from pyftpdlib.authorizers import DummyAuthorizer

authorizer = DummyAuthorizer()
authorizer.add_user('python', '123456', '/home/user', perm='elradfmwM')
handler = FTPHandler
handler.authorizer = authorizer

server = FTPServer(('0.0.0.0', 21111), handler)
server.serve_forever()
